@@ -1,5 +1,5 @@
 
-__version__ = "0.1.1"
+__version__ = "0.2.1"
 
 from mpgameserver.auth import Auth
 from mpgameserver.client import UdpClient
@@ -8,10 +8,17 @@ from mpgameserver.context import ServerContext
 from mpgameserver.crypto import EllipticCurvePrivateKey, EllipticCurvePublicKey
 from mpgameserver.handler import EventHandler
 from mpgameserver.logger import setupLogger
-from mpgameserver.serializable import SerializableType, Serializable, SerializableEnum
+from mpgameserver.serializable import SerializableType, Serializable, SerializableEnum, Default
 from mpgameserver.task import TaskPool
 from mpgameserver.timer import Timer
 from mpgameserver.twisted import TwistedServer, ThreadedServer
+from mpgameserver.http_server import path_join_safe, \
+    get, put, delete, post, \
+    Router, Resource, HTTPServer, \
+    Response, JsonResponse, SerializableResponse
+from mpgameserver.http_client import HTTPClient
+from mpgameserver.dispatch import ServerMessageDispatcher, ClientMessageDispatcher, \
+    server_event, client_event
 
 # pygame is an optional dependency
 try:
