@@ -63,7 +63,6 @@ import os
 import sys
 import time
 import struct
-import logging
 import select
 import random
 from typing import Callable, List
@@ -73,9 +72,9 @@ from . import crypto
 from .crypto import EllipticCurvePrivateKey, EllipticCurvePublicKey
 
 from .timer import Timer
-from .logger import PeerLogger
+from .logger import mplogger, PeerLogger
 from .context import ServerContext
-from collections import namedtuple
+from collections import namedtuple, defaultdict
 
 class PendingMessage(object):
     def __init__(self, seq, type, payload, callback, retry):
@@ -295,9 +294,21 @@ class PacketIdentifier(SerializableEnum):
     TO_CLIENT = b"FSOC"
 
 class PacketType(SerializableEnum):
-
     """
 
+    The CLIENT_HELLO, SERVER_HELLO, CHALLENGE_RESP types implement a
+    three way handshake that agree on an encryption key and verify both
+    sides are implementing the same protocol.
+
+    :attr UNKNOWN:        not used
+    :attr CLIENT_HELLO:   sent by the client to initiate a connection
+    :attr SERVER_HELLO:   server response to the hello
+    :attr CHALLENGE_RESP: client response to finialize the connection
+    :attr KEEP_ALIVE:     for packets automatically generated by a timeout
+    :attr DISCONNECT:     client or server packet indicating graceful disconnect
+    :attr APP:            user defined packet
+    :attr APP_FRAGMENT:   user defined packet was split into fragments
+
     """
     UNKNOWN          = 0x00  #
     CLIENT_HELLO     = 0x01  #
@@ -325,7 +336,7 @@ class PacketHeader(object):
     | 2 | seq num                | incrementing packet counter |
     | 2 | previous acked seq num | sequence number of the last
                                          received packet |
-    | 1 | packet type            | descrbies payload content |
+    | 1 | packet type            | describbes payload content |
     | 2 | length                 | number of bytes in the payload
                                          excluding any CRC or AES-GCM tag |
     | 1 | count                  | number of messages in this packet |
@@ -666,6 +677,7 @@ class HandshakeServerHelloMessage(Serializable):
     of the message.
 
     """
+    server_root_pubkey: object = None
     server_pubkey: object = None
     salt: bytes = b""
     token: int = 0
@@ -681,14 +693,26 @@ class HandshakeServerHelloMessage(Serializable):
 
         signature = kwargs['server_root_key'].sign(payload)
 
+        serialize_value(stream, kwargs['server_root_key'].getPublicKey().getBytes())
         serialize_value(stream, payload)
         serialize_value(stream, signature)
 
     def deserialize(self, stream, **kwargs):
+
+        root_pubkey = deserialize_value(stream, **kwargs)
+        self.server_root_pubkey = EllipticCurvePublicKey.fromBytes(root_pubkey)
+
         payload = deserialize_value(stream, **kwargs)
         signature = deserialize_value(stream, **kwargs)
 
-        kwargs['server_public_key'].verify(signature, payload)
+        # if no key is provided, use the key given by the server.
+        # Use a preshared key to avoid main in the middle attacks
+        if kwargs['server_public_key'] is None:
+            key = self.server_root_pubkey
+        else:
+            key = kwargs['server_public_key']
+
+        key.verify(signature, payload)
 
         temp = BytesIO(payload)
 
@@ -717,11 +741,13 @@ class ConnectionStatus(SerializableEnum):
     :attr CONNECTED: the client is connected, keys are set
     :attr DISCONNECTING: the client is closing the connection gracefully
     :attr DISCONNECTED: the client is not connected
+    :attr DROPPED: the client lost communication with the server
     """
     CONNECTING    = 1
     CONNECTED     = 2
     DISCONNECTING = 3
     DISCONNECTED  = 4
+    DROPPED       = 5
 
 class ConnectionQuality(SerializableEnum):
     EXCELLENT = 1
@@ -1092,7 +1118,7 @@ class ConnectionBase(object):
 
         pkt_type = PacketType.UNKNOWN
         msgs = [] # messages (seq, typ, msg) to include in this packet
-        msg_length = 0 # sum of length of messages in msgs, excluding overhead
+        current_msg_length = 0 # sum of length of messages in msgs, excluding overhead
 
         # resend any messages that had the resend flag set
         # and have not yet timed out or been acked. the resend_delay is a
@@ -1107,11 +1133,13 @@ class ConnectionBase(object):
                     idx += 1
                     continue
 
-                size = len(msg.payload) + Packet.overhead(1+len(msgs)) + msg_length
+                # calculate the size of the packet so far + this message
+                size = len(msg.payload) + Packet.overhead(1+len(msgs)) + current_msg_length
+                # if the message fits add it to the packet
                 if size <= Packet.MAX_PAYLOAD_SIZE:
                     del self.pending_retry_msg[msgseq]
                     msgs.append(msg)
-                    msg_length += len(msg.payload)
+                    current_msg_length += len(msg.payload)
 
                 idx += 1
 
@@ -1122,11 +1150,13 @@ class ConnectionBase(object):
         while idx < len(self.outgoing_messages):
             pending = self.outgoing_messages[idx]
 
-            size = len(pending.payload) + Packet.overhead(1+len(msgs)) + msg_length
+            # calculate the size of the packet so far + this message
+            size = len(pending.payload) + Packet.overhead(1+len(msgs)) + current_msg_length
+            # if the message fits add it to the packet
             if size <= Packet.MAX_PAYLOAD_SIZE:
-                msg = self.outgoing_messages.pop(idx)
-                msgs.append(msg)
-                msg_length += len(msg.payload)
+                self.outgoing_messages.pop(idx)
+                msgs.append(pending)
+                current_msg_length += len(pending.payload)
             else:
                 idx += 1
 
@@ -1225,7 +1255,7 @@ class ConnectionBase(object):
         try:
             pkt = Packet.from_bytes(hdr, self.session_key_bytes, datagram)
         except Exception as e:
-            self.log.exception("unable to decode packet %s", hdr)
+            #self.log.exception("unable to decode packet %s", hdr)
             self.stats.dropped += 1
             return False
 
@@ -1234,7 +1264,6 @@ class ConnectionBase(object):
             # if inserting dropped unacked bits then those packets will time out
             # the user may want to know to slow down the sending rate
             # and/or the protocol may want to back off.
-
             self.bitfield_pkt.insert(pkt.hdr.seq)
         except DuplicationError:
             self.stats.dropped += 1
@@ -1265,7 +1294,13 @@ class ConnectionBase(object):
         try:
             self.bitfield_msg.insert(msgseq)
         except DuplicationError:
-            print("drop duplicated seq", msgseq, Serializable.loadb(msg))
+
+            #try:
+            #    msgdat = Serializable.loadb(msg)
+            #    name = str(type(msgdat))
+            #except Exception as e:
+            #    name = msg[:16]
+            #print("drop duplicated typ=%s seq=%s (%s)" % (pkt_typ, msgseq, name))
             return
 
         if pkt_typ == PacketType.CLIENT_HELLO:
@@ -1443,6 +1478,9 @@ class ClientServerConnection(ConnectionBase):
             if len(self.stats.latency) > 5 * 60:
                 self.stats.latency.pop(0)
 
+        if self.last_recv_time > 0 and t0 > self.last_recv_time + 5:
+            self.status = ConnectionStatus.DROPPED
+
         if self.time_client_hello_sent and self.connection_callback:
             # TODO: its possible for the server hello to come in after the timeout
             #   solution: uid in client hello that must be returned
@@ -1454,8 +1492,8 @@ class ClientServerConnection(ConnectionBase):
 
     def _sendClientHello(self):
         self.log.debug("send client hello")
-        if self.server_public_key is None:
-            raise ProtocolError("no public key set")
+        #if self.server_public_key is None:
+        #    raise ProtocolError("no public key set")
 
         msg = HandshakeClientHelloMessage()
         msg.client_pubkey = self.session_key.getPublicKey()
@@ -1498,12 +1536,12 @@ class ClientServerConnection(ConnectionBase):
 
     def _ClientHelloTimeout(self, success):
         if not success:
-            logging.error("unable to connect to server")
+            mplogger.error("unable to connect to server")
 
     def _ChallengeResponseTimeout(self, success):
         if not success:
             # expecting a keep alive or other message
-            logging.error("no response to challenge")
+            mplogger.error("no response to challenge")
 
 class ServerClientConnection(ConnectionBase):
     """ Many of the events in EventHandler receive a client. That client implements
