@@ -67,16 +67,24 @@ import struct
 from typing import NewType, Generic, List, Dict, Tuple, get_args, get_origin
 from collections.abc import Iterable, Sequence, Mapping
 import sys
-import logging
 from io import BytesIO
 from enum import Enum
 import gzip
 
+from .logger import mplogger
+
 # maximum string or bytes length is 1mb.
 MAX_BYTES_LENGTH = 2**20
 
 MAX_ARRAY_LENGTH = 2**14
 
+def _default():
+    class Default(object):
+        def __repr__(self):
+            return "serializable.Default"
+    return Default()
+Default = _default()
+
 class SerializableBaseTypes(object):
 
     bool_t    = 1
@@ -150,7 +158,7 @@ class SerializableType(type):
         cls._fields = tuple(name for name in cls.__dict__ if ispublic(cls, name))
         for field in cls._fields:
             if field not in cls.__annotations__:
-                logging.info("missing annotation for %s.%s\n" % (cls.__name__, field))
+                mplogger.info("missing annotation for %s.%s\n" % (cls.__name__, field))
         return cls
 
     @staticmethod
@@ -229,7 +237,7 @@ serialize_types = {
     type(None): serialize_null,
 }
 
-def serialize_value(stream, value, **kwargs):
+def serialize_value(stream, value, _field=None, **kwargs):
 
     t = type(value)
 
@@ -251,7 +259,7 @@ def serialize_value(stream, value, **kwargs):
         value.serialize_header(stream)
         value.serialize(stream, **kwargs)
     else:
-        raise TypeError("%s" % type(value))
+        raise TypeError("%s: %s" % (_field, type(value)))
 
 def serialize_map(stream, value):
     if len(value) > MAX_ARRAY_LENGTH:
@@ -298,18 +306,18 @@ class SerializableError(Exception):
 class SerializableHeaderError(SerializableError):
     pass
 
-deserialize_bool_t    = lambda stream: struct.unpack(">?", stream.read(1))[0]
-deserialize_int8_t    = lambda stream: struct.unpack(">b", stream.read(1))[0]
-deserialize_int16_t   = lambda stream: struct.unpack(">h", stream.read(2))[0]
-deserialize_int32_t   = lambda stream: struct.unpack(">l", stream.read(4))[0]
-deserialize_int64_t   = lambda stream: struct.unpack(">q", stream.read(8))[0]
-deserialize_uint8_t   = lambda stream: struct.unpack(">B", stream.read(1))[0]
-deserialize_uint16_t  = lambda stream: struct.unpack(">H", stream.read(2))[0]
-deserialize_uint32_t  = lambda stream: struct.unpack(">L", stream.read(4))[0]
-deserialize_uint64_t  = lambda stream: struct.unpack(">Q", stream.read(8))[0]
-deserialize_float32_t = lambda stream: struct.unpack(">f", stream.read(4))[0]
-deserialize_float64_t = lambda stream: struct.unpack(">d", stream.read(8))[0]
-deserialize_null_t    = lambda stream: None
+deserialize_bool_t    = lambda stream, **kwargs: struct.unpack(">?", stream.read(1))[0]
+deserialize_int8_t    = lambda stream, **kwargs: struct.unpack(">b", stream.read(1))[0]
+deserialize_int16_t   = lambda stream, **kwargs: struct.unpack(">h", stream.read(2))[0]
+deserialize_int32_t   = lambda stream, **kwargs: struct.unpack(">l", stream.read(4))[0]
+deserialize_int64_t   = lambda stream, **kwargs: struct.unpack(">q", stream.read(8))[0]
+deserialize_uint8_t   = lambda stream, **kwargs: struct.unpack(">B", stream.read(1))[0]
+deserialize_uint16_t  = lambda stream, **kwargs: struct.unpack(">H", stream.read(2))[0]
+deserialize_uint32_t  = lambda stream, **kwargs: struct.unpack(">L", stream.read(4))[0]
+deserialize_uint64_t  = lambda stream, **kwargs: struct.unpack(">Q", stream.read(8))[0]
+deserialize_float32_t = lambda stream, **kwargs: struct.unpack(">f", stream.read(4))[0]
+deserialize_float64_t = lambda stream, **kwargs: struct.unpack(">d", stream.read(8))[0]
+deserialize_null_t    = lambda stream, **kwargs: None
 
 deserialize_types = {
     SerializableBaseTypes.bool_t:      deserialize_bool_t,
@@ -329,8 +337,8 @@ deserialize_types = {
 def deserialize_value(stream, **kwargs):
 
     """
-    todo: allow passing a custom registry in as a kwarg
-          for non-builtin types use the registry when decoding
+    kwargs can contain a 'registry', a mapping of type_id to Serializable
+    class names.
 
     when serializing to persistent storage, store the registry
     as a sequence of typeid: string-name (using basic types)
@@ -342,11 +350,7 @@ def deserialize_value(stream, **kwargs):
         raise SerializableHeaderError("unexpected end of stream")
     type_id, = struct.unpack(">H", buf)
 
-    if 'registry' in kwargs:
-        registry = kwargs['registry']
-    else:
-        registry = SerializableType.registry
-
+    registry = kwargs.get('registry', SerializableType.registry)
 
     if type_id in deserialize_types:
         typ_ = deserialize_types[type_id]
@@ -357,7 +361,7 @@ def deserialize_value(stream, **kwargs):
 
     try:
         if type_id in deserialize_types:
-            return deserialize_types[type_id](stream)
+            return deserialize_types[type_id](stream, **kwargs)
         elif type_id in registry:
             obj = registry[type_id]()
             obj.deserialize(stream, **kwargs)
@@ -369,9 +373,9 @@ def deserialize_value(stream, **kwargs):
         # raise a new error with the last type that successfully decoded
         raise SerializableError("unable to deserialize type %s" % typ_)
 
-def deserialize_string(stream):
+def deserialize_string(stream, **kwargs):
 
-    length = deserialize_value(stream)
+    length = deserialize_value(stream, **kwargs)
     if not isinstance(length, int):
         raise TypeError("expected integer length, found %s" % type(length))
 
@@ -380,9 +384,9 @@ def deserialize_string(stream):
 
     return stream.read(length).decode("utf-8")
 
-def deserialize_bytes(stream):
+def deserialize_bytes(stream, **kwargs):
 
-    length = deserialize_value(stream)
+    length = deserialize_value(stream, **kwargs)
 
     if not isinstance(length, int):
         raise TypeError("expected integer length, found %s" % type(length))
@@ -392,8 +396,8 @@ def deserialize_bytes(stream):
 
     return stream.read(length)
 
-def deserialize_map(stream):
-    length = deserialize_value(stream)
+def deserialize_map(stream, **kwargs):
+    length = deserialize_value(stream, **kwargs)
 
     if not isinstance(length, int):
         raise TypeError("expected integer length, found %s" % type(length))
@@ -403,14 +407,14 @@ def deserialize_map(stream):
 
     obj = {}
     for i in range(length):
-        k = deserialize_value(stream)
-        v = deserialize_value(stream)
+        k = deserialize_value(stream, **kwargs)
+        v = deserialize_value(stream, **kwargs)
         obj[k] = v
 
     return obj
 
-def deserialize_seq(stream):
-    length = deserialize_value(stream)
+def deserialize_seq(stream, **kwargs):
+    length = deserialize_value(stream, **kwargs)
 
     if not isinstance(length, int):
         raise TypeError("expected integer length, found %s" % type(length))
@@ -422,15 +426,15 @@ def deserialize_seq(stream):
     for i in range(length):
 
         try:
-            o = deserialize_value(stream)
+            o = deserialize_value(stream, **kwargs)
             obj.append(o)
         except Exception as e:
             raise e
 
     return obj
 
-def deserialize_set(stream):
-    length = deserialize_value(stream)
+def deserialize_set(stream, **kwargs):
+    length = deserialize_value(stream, **kwargs)
 
     if not isinstance(length, int):
         raise TypeError("expected integer length, found %s" % type(length))
@@ -438,7 +442,7 @@ def deserialize_set(stream):
     if length > MAX_ARRAY_LENGTH:
         raise ValueError("set length too large: %d" % length)
 
-    obj = set([deserialize_value(stream) for i in range(length)])
+    obj = set([deserialize_value(stream, **kwargs) for i in range(length)])
 
     return obj
 
@@ -457,7 +461,8 @@ def deserialize_registry(stream, **kwargs):
         type_id = deserialize_value(stream, **kwargs)
         cls_name = deserialize_value(stream, **kwargs)
 
-        obj[type_id] = SerializableType.names[cls_name]
+        if cls_name in SerializableType.names:
+            obj[type_id] = SerializableType.names[cls_name]
     return obj
 
 def _fromJsonBasic(type, field, value):
@@ -500,12 +505,16 @@ class Serializable(object, metaclass=SerializableType):
     """
 
     def __init__(self, **kwargs):
+
         for attr, t in self.__annotations__.items():
             if isinstance(t, Serializable) or \
                t in (dict, list, set, tuple):
                 setattr(self, attr, t())
             elif get_origin(t) in (dict, list, set, tuple):
                 setattr(self, attr, get_origin(t)())
+            else:
+                if getattr(self, attr) is Default:
+                    setattr(self, attr, t())
 
         for k,v in kwargs.items():
             if k in self._fields:
@@ -534,8 +543,9 @@ class Serializable(object, metaclass=SerializableType):
 
         """
         #stream.write(struct.pack(">H", self.type_id))
+        serialize_value(stream, len(self._fields))
         for field in self._fields:
-            serialize_value(stream, getattr(self, field))
+            serialize_value(stream, getattr(self, field), field)
 
     def deserialize(self, stream, **kwargs):
         """populate member attributes by reading fields from a stream.
@@ -544,11 +554,31 @@ class Serializable(object, metaclass=SerializableType):
         :param kwargs: extra arguments that are passed to deserialize, a specialization of deserialize
         can use kwargs to control how the class is deserialized
 
+        :returns: self
+
+        Note: the return value can be any serializable type
+        to implement versioning, old versions can be deserialized and then
+        convert to the new version and return that instead by
+        reimplementing this function for that type.
+
         """
-        for field in self._fields:
-            setattr(self, field, deserialize_value(stream))
+        #for field in self._fields:
+        num_fields = deserialize_value(stream, **kwargs)
+        for i in range(num_fields):
+            field = self._fields[i]
+            setattr(self, field, deserialize_value(stream, **kwargs))
         return self
 
+    def store_persistant(self, stream):
+        """ private write to long term storage
+
+        write a registry from the stream, which includes the mapping
+        of type_id to message types. then serialize the message to the stream
+        """
+
+        serialize_registry(stream)
+        serialize_value(stream, self)
+
     def dumpb(self, **kwargs):
         """ return a byte array representation of this class
 
@@ -557,7 +587,8 @@ class Serializable(object, metaclass=SerializableType):
 
         """
         stream = BytesIO()
-        stream.write(struct.pack(">H", self.type_id))
+        #stream.write(struct.pack(">H", self.type_id))
+        self.serialize_header(stream)
         self.serialize(stream, **kwargs)
         return stream.getvalue()
 
@@ -593,6 +624,19 @@ class Serializable(object, metaclass=SerializableType):
     def loads(cls, string, **kwargs):
         return cls.fromJson(json.loads(string, **kwargs))
 
+    @staticmethod
+    def load_persistant(stream, **kwargs):
+        """ private load from long term storage
+
+        read a registry from the stream, which includes the mapping
+        of type_id to message types. use the registry to deserialize
+        the rest of the stream
+        """
+        if isinstance(stream, bytes):
+            stream = BytesIO(stream)
+        registry = deserialize_registry(stream)
+        return deserialize_value(stream, registry=registry)
+
     @classmethod
     def fromJson(cls, record):
         """ initialize a new instance of Type from a JSON object
@@ -673,7 +717,6 @@ class Serializable(object, metaclass=SerializableType):
         In addition, no conversion is done for float types. This means that invalid
         JSON float values are not modified.
         """
-
         obj = {}
         for field in self._fields:
 
@@ -718,8 +761,9 @@ class Serializable(object, metaclass=SerializableType):
                     raise TypeError("%s != %s" % (origin, type(record)))
 
             else:
-                obj[field] = _toJsonBasic(self.__annotations__[field],
-                    field, getattr(self, field))
+                type_ = self.__annotations__[field]
+                value = getattr(self, field)
+                obj[field] = _toJsonBasic(type_, field, value)
         return obj
 
 class SerializableEnumType(type):
@@ -774,7 +818,6 @@ class SerializableEnum(object, metaclass=SerializableEnumType):
         elif value is None:
             self.value = value
         else:
-            self.value = value
             raise ValueError(value)
 
     def __repr__(self):
@@ -817,7 +860,7 @@ class SerializableEnum(object, metaclass=SerializableEnumType):
 
     @classmethod
     def fromJson(cls, value):
-        return cls._name2value[value]
+        return cls._name2value[value.upper()]
 
     def toJson(self):
         """ return a dictionary suitable for passing to json.dumps
@@ -848,8 +891,50 @@ class SerializableEnum(object, metaclass=SerializableEnumType):
     def __bool__(self):
         return bool(self.value)
 
+
 def main():  #pragma: no cover
-    pass
+
+
+    class Color(SerializableEnum):
+        RED=1
+        BLUE=2
+        GREEN=3
+
+    class Shape(SerializableEnum):
+        SQUARE=1
+        RECTANGLE=2
+        CIRCLE=3
+
+    class Block(Serializable):
+        color: Color = Color.RED
+        shape: Shape = Shape.SQUARE
+
+    class Block2(Serializable):
+        block: Block = Default
+
+    block = Block()
+    print(block.toJson())
+    print(Block.fromJson({"color": "BLUE"}))
+
+
+    block2 = Block2()
+    print(block2)
+
+    print(Block._fields)
+    print(block)
+    print(block.toJson())
+
+    stream = BytesIO()
+    block.store_persistant(stream)
+    print(stream.getvalue())
+
+    stream.seek(0)
+    block2 = Serializable.load_persistant(stream)
+    print(block2)
+
+
+
+
 
 if __name__ == '__main__':  #pragma: no cover
     main()
\ No newline at end of file
